/* This file is part of the software group
   Copyright (C) 2010 Alain Le Bot
   Laboratoire de tribologie et dynamique des systemes CNRS
   Ecole centrale de Lyon
   36, av. Guy de Collongue 69134 Ecully FRANCE.

   All rights reserved. Tous droits reserves.

   File group.c
   A. Le Bot le 3 january 2010 

   Last modification 3 ojanuary 2010. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <getopt.h>
#include <math.h>
#include <limits.h>

#define NAME "group"
#define VERSION "0.1"
#define DATE "3 january 2010"

#define ORDER 50                                   // order maximum of group
#define SYMBOL 4                                   // size of element symbols
#define TEXT_LENGTH ORDER*(ORDER+1)*(SYMBOL)+25
#define KEYWORD_LENGTH 15

//int check_commutativity(int*, int);
//int check_associativity(int*, int);
//int check_unity(int*, int, int*);
//int check_invertibility(int*, int*, int, int);
//int read_data(char*, char*, char*, char*, int*);
void getField(char*, char*, int, void*);
int sgets(char*, int, char*);
//int print_numeric_table(int*, int);
//int print_symbolic_table(char*, char*, char*, int);
//int print_inverse(char*, int*, int);
//int translate_data(int, char*, char*, int*);

struct option longopts[] = {
  {"help", no_argument, NULL, 'h'},
  {"version", no_argument, NULL, 'v'},
  {"structure", no_argument, NULL, 's'},
  {NULL, 0, NULL, 0}
};

static char *progname;
int flagh = 0, flagv = 0, flags = 0, lose = 0;

void main(int argc, char *argv[])
{
 char *filename;
 int  optc;
 int  order, unit;
 char set[ORDER][SYMBOL];
 char table[ORDER][ORDER][SYMBOL];
 char operation[SYMBOL];
 int  commutativity, associativity, invertibility, unity;
 int  mult[ORDER][ORDER],inverse[ORDER];
 int  i,j;

  progname = argv[0];

  while ((optc =
	  getopt_long (argc, argv, "hvs", longopts,
		       (int *) 0)) != -1)
    switch (optc)
      {
      case 'h':
	flagh = 1;
	break;
      case 'v':
	flagv = 1;
	break;
      case 's':
	flags = 1;
	break;
      default:
	lose = 1;
	break;
      }

/* invalid argument */
  if (lose)
    {
      fprintf (stderr, "Invalid option\n");
      fprintf (stderr, "Try `%s --help' for more information.\n", progname);
      exit (-1);
    }

/* get file name */
  if (!flagv&&!flagh)
    {
     optind++;
     if (optind>argc)
       {
        fprintf  (stderr, "Too few arguments\n");
        fprintf  (stderr, "Try `%s --help' for more information.\n", progname);
        exit (-1);
       }
     filename=argv[optind-1];
    }

/* too many arguments */
  if (optind < argc)
    {
      fprintf (stderr, "Too many arguments\n");
      fprintf (stderr, "Try `%s --help' for more information.\n", progname);
      exit (-1);
    }

/* help */
  if (flagh)
    {
      printf ("%s, basic properties of finite groups.\n\n",progname);
      printf ("Usage: %s [OPTION] [FILE]\n\n", progname);
      printf ("\
  -h, --help      display this help and exit\n\
  -v, --version   display version information and exit\n\
  -s, --structure test basic features of operation\n\n");
      printf ("Report bugs to <alain.le-bot@ec-lyon.fr>.\n");

      exit (0);
    }

/* version */
  if (flagv)
    {
      printf ("%s: code to determine the basic properties of finite groups\n", NAME);
      printf ("Version %s Last modification %s\n\n", VERSION,DATE);
      printf ("This software tests the basic properties of finite groups. The group is defined\n");
      printf ("by giving the operation table in input file. Then group can test associativity,\n");
      printf ("commutativity, invertibility of elements and gives the unit element.\n\n");
      printf ("Copyright (C) 2010 Alain Le Bot, all rights reserved.\n\n");
      printf ("Bug reports to alain.le-bot@ec-lyon.fr\n\n");

      exit (0);
    }

/* Read data file */
 read_data(filename,set,table,operation,&order);
 print_symbolic_table(set,table,operation,order);

/* Check basic properties */
 if (translate_data(order,set,table,mult)!=0)
  {
   printf("Closure axiom violated\n");
   exit(0);
  }
// print_numeric_table(mult,order);
 associativity=(check_associativity(mult,order)==0);
 if (associativity!=1)
  {
   printf("Associativity axiom violated\n");
   exit(0);
  }
 unity=(check_unity(mult,order,&unit)==0);
 if (unity!=1)
  {
   printf("No unit element found\n");
   exit(0);
  }
 invertibility=(check_invertibility(mult,inverse,order,unit)==0);
 commutativity=(check_commutativity(mult,order)==0);

 if (associativity&&commutativity&&unity&&invertibility)
  {
   print_inverse(set, inverse, order); 
   printf("The set is an Abelian group\n\n");
  }
 else if (associativity&&unity&&invertibility)
  {
   print_inverse(set, inverse, order);
   printf("The set is a non Abelian group\n\n");
  }
 else if (associativity&&unity)
  printf("The set is a monoid\n\n");
 else
  printf("The set is nothing\n\n");

/* Find subgroups */
find_subgroup(set, mult, inverse, order);

 exit(0);

}

/*cccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!         Check commutativity of operation
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
int check_commutativity(int mult[][ORDER], int n)
{
 int a,b;

 for (a=0; a<n; a++)
  {
   for (b=0; b<n; b++)
    {
     if (mult[a][b]!=mult[b][a]) return -1;
    }
  }
 return 0;

}



/*cccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!         Check associativity of operation
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
int check_associativity(int mult[][ORDER], int n)
{
 int a,b,c,ab,bc;

 for (a=0; a<n; a++)
  {
   for (b=0; b<n; b++)
    {
     for (c=0; c<n; c++)
      {
       ab=mult[a][b];
       bc=mult[b][c];
       if (mult[a][bc]!=mult[ab][c]) return -1;
      }
    }
  }
 return 0;

}



/*cccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!         Seek unit element of multiplication
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
int check_unity(int mult[][ORDER], int n, int* e)
{
 int a,b;

 for (a=0; a<n; a++)
  {
   for (b=0; b<n; b++)
     if (mult[a][b]!=b||mult[b][a]!=b) break;
   if (b==n) 
    {
     *e=a;
     return 0;
    }
  }
 return -1;
}



/*cccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!    Check the existence of symmetric for all elements
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
int check_invertibility(int mult[][ORDER], int inverse[], int n, int e)
{
 int a,b;

 for (a=0; a<n; a++)
   for (b=0; b<n; b++)
     if ((mult[a][b]==e)&&(mult[b][a]==e)) 
      {
       inverse[a]=b;
       break;
      }
   
 if (a==n) 
  return 0;
 else 
  return -1;

}

/*cccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!    Check if a subset is a subgroup
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
int check_subgroup(int mult[][ORDER], int inverse[], int subgroup[], int p)
{
 int a,b,c,i,j,k;

 for (i=0; i<p; i++)
  {
   a=subgroup[i];
   for (j=0; j<p; j++)
    {
     b=inverse[subgroup[j]];
     c=mult[a][b];
     for (k=0; k<p; k++)
       if (c==subgroup[k]) break;
     if (k==p) return -1;
    }
  }
 
 return 0;

}

/*cccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!    Check if a subgroup is normal
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
int check_normality(int mult[][ORDER], int inverse[], int subgroup[], int p, int n)
{
 int a,b,c,i,j,k;

 for (a=0; a<n; a++)
 for (i=0; i<p; i++)
  {
   b=mult[a][subgroup[i]];
   b=mult[b][inverse[a]];
   for (j=0; j<p; j++)
    {
     c=subgroup[j];
     if (b==c) break;
    }
   if (j==p) return -1;
  }
 
 return 0;

}

/* Print the subset */
void print_subset(char set[][SYMBOL], int subset[], int p, char label[]) {
 int i;
 printf("{");
 for (i = 0; i < p; i++)
   printf("%s,", set[subset[i]]);
 printf("\b} %s\n",label); // \b to suppress the last coma
}

/* Generates the next mask */
int next(int mask[], int n) {
	int i;
	for (i = 0; (i < n) && mask[i]; ++i)
		mask[i] = 0;

	if (i < n) {
		mask[i] = 1;
		return 1;
	}
	return 0;
}

int find_subgroup(char set[][SYMBOL], int mult[][ORDER], int inverse[], int n)
{
	int subset[ORDER];
        int mask[ORDER];
	int i,j,p;

        printf("List of subgroups\n");

	for (i = 0; i < n; i++)
	 mask[i] = 0;

	while (next(mask, n))
         {
          for (i=0, p=0; i<n; i++)
           if (mask[i]) subset[p++]=i;
          if (!check_subgroup(mult, inverse, subset, p))
           {
            if (!check_normality(mult, inverse, subset, p, n))
             print_subset(set, subset, p,"NORMAL");
            else
             print_subset(set, subset, p,"");
           }
         }
/*
        for (i=1; i<(1<<n); i++)       // The number of subsets is (1<<n)=pow(2,n)
         {
          for (j=0; j<n; j++)
           mask[j]=(i>>j)%2;           // translate binary mask j to integer mask
          for (j=0, p=0; j<n; j++)
           if (mask[j]) subset[p++]=j; // translate integer mask to list of element
          if (!check_subgroup(mult, inverse, subset, p))
           print_subset(set, subset, p);
         }
*/
	return 0;

}

/*cccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!                    Read data file
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
int read_data(char* filename, char set[][SYMBOL], char table[][ORDER][SYMBOL], char operation[], int* n)
{
   FILE *fp;
   int j;
   int   i=-1, comment=0, blank=0;
   char tmp, text[TEXT_LENGTH];

   if ((fp=fopen(filename,"r"))==NULL)
    {
     printf("ERROR: Unable to open file %s\n",filename);
     exit(-1);
    }

   while ((tmp=getc(fp))!=EOF&&i<TEXT_LENGTH)
    {
     if (comment==0&&tmp=='#')  {text[i++]=' '; comment=1;}  // Be carefull to the case where # is the last character of a line
     if (comment==0&&tmp=='\n') tmp=' ';
     if (comment==0&&tmp=='\t') tmp=' ';
     if (comment==0&&tmp=='\r') tmp=' ';
     if (comment==0&&tmp!=' ')  blank=0;
     if (comment==0&&blank==0)  text[i++]=tmp;
     if (comment==0&&tmp==' ')  blank=1;
     if (comment==1&&tmp=='\n') comment=0;
    }
   text[i]='\0';

//   printf("file=%s\n",text);

   getField(text,"ORDER",1,n);
   if (*n>=ORDER)
    {
     printf("ERROR: too large set\n");
     exit(-1);
    }
   getField(text,"OPERATION",1,operation);
   getField(text,"SET",*n,set);
   getField(text,"TABLE",*n,table);

   close(fp);

}

// copy the value of field word in field
void getField(char* keyword, char* word, int n, void* field)
{
 char   pattern[KEYWORD_LENGTH];
 char*  ptrField;
 int i,j;

 strcpy(pattern,word);
 strcat(pattern,"=");

 if ((ptrField=strstr(keyword,pattern))==NULL)
 {
  printf("ERROR: field %s not found\n",word);
  exit(-1);
 }

 ptrField=ptrField+strlen(pattern); // shift after pattern

 if (!strcmp(word,"ORDER"))
  {
   if (sscanf(ptrField,"%d", (int*)   field)!=1) goto errorGetField;
  }
 if (!strcmp(word,"OPERATION"))
  {
   sgets((char*) field,1,ptrField);
  }
 if (!strcmp(word,"SET"))
  {
   for (i=0; i<n; i++)
     ptrField=ptrField+sgets((char*) field+SYMBOL*i,SYMBOL-1,ptrField);
  }
 if (!strcmp(word,"TABLE"))
  {
   for (i=0; i<n; i++)
   for (j=0; j<n; j++)
     ptrField=ptrField+sgets((char*) field+SYMBOL*ORDER*i+SYMBOL*j,SYMBOL-1,ptrField);
  }
 return;

 errorGetField:
 printf("ERROR: unable to read %s value\n",word);
 exit(-1);

}

// read next string in chaine2 (delimited by blanks) and write it in chaine 1 with a final '\0'
// max is the max number of character to read (without final '\0')
// return the number of characters read in chaine 2 (including blanks)
int sgets(char* chaine1, int max, char* chaine2)
{
 int i,j;
 i=0;
 j=0;
 while (chaine2[j]==' ') j++;
 while (chaine2[j]!=' '&&chaine2[j]!='\0'&&i<max)
  {
   chaine1[i++]=chaine2[j++];
  }
 chaine1[i]='\0';
 return j;
}

/*cccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!                  Print operation table
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
int print_numeric_table(int mult[][ORDER], int n)
{
   int a,b;

   printf("Numeric table\n");
   printf(" + ");
   for (a=0; a<n; a++)
    printf("%2d ",a);
   printf("\n");
   for (a=0; a<n; a++)
    {
     printf("%2d ",a);
     for (b=0; b<n; b++)
      printf("%2d ",mult[a][b]);
     printf("\n");
    }

}

int print_symbolic_table(char set[][SYMBOL], char table[][ORDER][SYMBOL], char operation[], int n)
{
   int a,b;

   printf("Operation table\n");
   printf("  %1s ",operation);
   for (a=0; a<n; a++)
    printf("%3s ",set[a]);
   printf("\n");
   for (a=0; a<n; a++)
    {
     printf("%3s ",set[a]);
     for (b=0; b<n; b++)
      printf("%3s ",table[a][b]);
     printf("\n");
    }
   printf("\n");

}

int print_inverse(char set[][SYMBOL], int inverse[], int n)
{
   int a,b;

   printf("Inverse table\n");
   printf("element ");
   for (a=0; a<n; a++)
    printf("%3s ",set[a]);
   printf("\n");
   printf("inverse ");
   for (a=0; a<n; a++)
    printf("%3s ",set[inverse[a]]);
   printf("\n\n");

}
/*cccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!                  Translate data
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
int translate_data(int n, char set[][SYMBOL], char table[][ORDER][SYMBOL], int mult[][ORDER])
{
   int a,b,i;

   for (a=0; a<n; a++)
   for (b=0; b<n; b++)
    {
     for (i=0; i<n; i++)
      if (!strcmp(table[a][b],set[i]))
       {
        mult[a][b]=i;
        break;
       }
     if (i==n) return -1;
    }
   return 0;
}

